/* pwcheck_ldap.c -- check passwords using LDAP
 *
 * Author: Clayton Donley <donley@cig.mot.com>
 *         http://www.wwa.com/~donley/
 * Version: 1.01
 *
 * Note: This works by finding a DN that matches an entered UID and
 * binding to the LDAP server using this UID.  This uses clear-text
 * passwords.  A better approach with servers that support SSL and
 * new LDAPv3 servers that support SASL bind methods like CRAM-MD5
 * and TSL.
 *
 * This version should work with both University of Michigan and Netscape
 * LDAP libraries.  It also gets rid of the requirement for userPassword
 * attribute readability.
 *
 *
 * Minor changes by Simon Loader <simon@surf.org.uk>
 *	Mainly to do with making it slightly more configurable and
 * some memory leaks to do with openLdap 1.1.2 Version 
 *  NOT TESTED AGAINST NETSCAPE.
 *
 * Changes by Andy Igoshin <ai@vsu.ru>
 */

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <lber.h>
#include <ldap.h>

/* Set These to your Local Environment */

#ifndef MY_LDAP_SERVER
#define MY_LDAP_SERVER	"127.0.0.1"
#endif
#ifndef MY_LDAP_BASEDN
#define MY_LDAP_BASEDN	"o=JOFA, c=UK"
#endif
#ifndef MY_LDAP_UIDATTR
#define MY_LDAP_UIDATTR	"uid"
#endif

int pwcheck(char *userid, char *password,
            char *my_ldap_server, char *my_ldap_basedn)
{
    LDAP *ld;
    LDAPMessage *result;
    LDAPMessage *entry;
    char *attrs[2];
    char filter[200]; 
    char *dn;

/* If the password is NULL, reject the login...Otherwise the bind will
   succeed as a reference bind.  Not good... */

    if (strcmp(password,"") == 0)
    {
       return 0;
    }

/* Open the LDAP connection.  Change the second argument if your LDAP
   server is not on port 389. */

    if ((ld = ldap_open(my_ldap_server,LDAP_PORT)) == NULL)
    {
	/* could contact it something is wrong */
       return 0;
    }

/*  Bind anonymously so that you can find the DN of the appropriate user.  */

    if (ldap_simple_bind_s(ld,"","") != LDAP_SUCCESS)
    {
        ldap_unbind(ld);
        return 0;
    }

/* Generate a filter that will return the entry with a matching UID */

    snprintf(filter,sizeof(filter),"(%s=%s)",MY_LDAP_UIDATTR,userid);

/* Just return country...This doesn't actually matter, since we will
   not read the attributes and values, only the DN */

    attrs[0] = "c";
    attrs[1] = NULL;

/* Perform the search... */

    if (ldap_search_s(ld,my_ldap_basedn,LDAP_SCOPE_SUBTREE,filter,attrs,1,&result) != LDAP_SUCCESS)
    {
       ldap_unbind(ld);
	/* search filed for some reason */
       return 0;
    }

/* If the entry count is not equal to one, either the UID was not unique or
   there was no match */

    if ((ldap_count_entries(ld,result)) != 1)
    {
		ldap_msgfree(result);
       ldap_unbind(ld);
	/* could find uid*/
       return 0;
    }

/* Get the first entry */

    if ((entry = ldap_first_entry(ld,result)) == NULL)
    {
		ldap_msgfree(result);
       ldap_unbind(ld);
	/*This should not fail unless the last condition failed */
       return 0;
    }

/* Get the DN of the entry */

    if ((dn = ldap_get_dn(ld,entry)) == NULL)
    {
		ldap_msgfree(entry);
       ldap_unbind(ld);
	/* couldnt find dn */
       return 0;
    }

/* Now bind as the DN with the password supplied earlier...
   Successful bind means the password was correct, otherwise the
   password is invalid. */

    if (ldap_simple_bind_s(ld,dn,password) != LDAP_SUCCESS)
    {
		free(dn);
		ldap_msgfree(entry);
       ldap_unbind(ld);
	/* couldnt bind therefore username or password wrong */	
       return 0;
    }

	free(dn);
	ldap_msgfree(entry);
    ldap_unbind(ld);
    return 1;
}
